import esm
import torch
import torch.nn.functional as F

from torch import nn, einsum
from einops import rearrange

class LightAttention(nn.Module):
    def __init__(self):
        super().__init__()
        self.att_conv = nn.Conv1d(1, 1, 9, 1, padding='same')
        self.feat_conv = nn.Conv1d(1, 1, 9, 1, padding='same')
    
    def forward(self, x):
        return self.att_conv(x).softmax(dim=-1) * self.feat_conv(x)

class Abyssal(nn.Module):
    def __init__(self):
        super().__init__()
        
        self.light_attention = LightAttention()
        self.fc_block = nn.Sequential(
            nn.Linear(2560, 2048),
            nn.ReLU(),  # Not sure
            nn.Linear(2048, 1024),
            nn.ReLU(),  # Not sure
            nn.Linear(1024, 1),
        )

    def forward(self, wt_emb, mut_emb):
        x_wt, x_mut = map(lambda t: rearrange(t, 'b h -> b 1 h'), [wt_emb, mut_emb])
        x_wt, x_mut = map(self.light_attention, [x_wt, x_mut])
        x = torch.cat([x_wt, x_mut], dim=-1)
        x = rearrange(x, 'b c h -> b (c h)')

        return self.fc_block(x)

if __name__ == '__main__':
    wt_emb = torch.randn([2, 1280])
    mut_emb = torch.randn([2, 1280])

    model = Abyssal()
    x = model(wt_emb, mut_emb)

    print(x.shape)



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import numpy as np
import random
import tqdm

from torch.utils.data import Dataset, DataLoader
from scipy.stats import pearsonr, spearmanr

from abyssal_pytorch import Abyssal
from abyssal_pytorch.data import MegaDataset

def train(model, train_loader, optimizer, criterion, metrics_f):
    model.train()
    running_output, running_label = [], []

    # Training loop with progressbar.
    bar = tqdm.tqdm(train_loader, total=len(train_loader), leave=False)
    for idx, batch in enumerate(bar):
        wt_emb, mut_emb = batch['wt_emb'].cuda(), batch['mut_emb'].cuda()
        label = batch['label'].cuda().flatten()

        optimizer.zero_grad()
        output = model(wt_emb, mut_emb).flatten()
        loss = criterion(output, label)
        loss.backward()
        optimizer.step()

        running_output.append(output.detach().cpu())
        running_label.append(label.detach().cpu())

        if idx % 100 == 0:
            running_output = torch.cat(running_output, dim=0)
            running_label = torch.cat(running_label, dim=0)

            running_loss = criterion(running_output, running_label)
            running_metrics = {k: f(running_output, running_label) for k, f in metrics_f.items()}

            loss = running_loss.item()
            pearson = running_metrics['pearson']
            spearman = running_metrics['spearman']
            bar.set_postfix(loss=loss, pearson=pearson, spearman=spearman)

            running_output, running_label = [], []

def validate(model, val_loader, criterion, metrics_f):
    model.eval()

    out, label = [], []
    with torch.no_grad():
        for idx, batch in enumerate(val_loader):
            wt_emb, mut_emb = batch['wt_emb'].cuda(), batch['mut_emb'].cuda()
            _label = batch['label'].cuda().flatten()

            _out = model(wt_emb, mut_emb).flatten()
            out.append(_out.cpu())
            label.append(_label.cpu())
        
    out = torch.cat(out, dim=0)
    label = torch.cat(label, dim=0)

    loss = criterion(out, label)
    metrics = {k: f(out, label) for k, f in metrics_f.items()}

    return loss, metrics

def seed_everything(seed):
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)

    # Performance drops, so commenting out for now.
    # torch.backends.cudnn.benchmark = False
    # torch.backends.cudnn.deterministic = True

def freeze(model):
    for param in model.parameters():
        param.requires_grad = False

def main():
    import pandas as pd
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--train', required=True)
    parser.add_argument('--val', required=True)
    parser.add_argument('--emb-dir', required=True)
    parser.add_argument('--batch-size', type=int, default=128)
    parser.add_argument('--epochs', type=int, default=72)
    parser.add_argument('--lr', type=float, default=1e-4)
    parser.add_argument('--seed', type=int, default=42)
    args = parser.parse_args()

    seed_everything(args.seed)

    train_df = pd.read_csv(args.train)
    train_set = MegaDataset(train_df, emb_dir=args.emb_dir)

    val_df = pd.read_csv(args.val)
    val_set = MegaDataset(val_df, emb_dir=args.emb_dir)

    train_loader = DataLoader(train_set, batch_size=args.batch_size, shuffle=True, drop_last=True, num_workers=16, pin_memory=True)
    val_loader = DataLoader(val_set, batch_size=args.batch_size, shuffle=False, drop_last=False, num_workers=16, pin_memory=True)

    model = Abyssal()
    model = model.cuda()

    optimizer = optim.Adam(model.parameters(), lr=args.lr)
    scheduler = optim.lr_scheduler.StepLR(optimizer, step_size=1, gamma=0.97)
    criterion = nn.MSELoss()

    metrics_f = {
        'pearson': lambda x, y: pearsonr(x, y)[0],
        'spearman': lambda x, y: spearmanr(x, y)[0],
    }

    for epoch in range(args.epochs):
        train(model, train_loader, optimizer, criterion, metrics_f)
        val_loss, val_metrics = validate(model, val_loader, criterion, metrics_f)

        print(f'Epoch {epoch}: val loss {val_loss:.4f}, val pearson {val_metrics["pearson"]:.4f}, val spearman {val_metrics["spearman"]:.4f}')
        scheduler.step()

if __name__ == '__main__':
    main()



import torch
import torch.nn as nn
import torch.nn.functional as F

from torch.utils.data import Dataset, DataLoader
from tqdm import tqdm
import numpy as np

class MegaDataset(Dataset):
    def __init__(self, df, emb_dir):
        """df: pandas dataframe for MegaTrain/MegaValidation/MegaTest datasets.
        """
        self.df = df.to_records()
        self.emb_dir = emb_dir

        print('Preloading embeddings..')
        wt_embs, mut_embs = [], []
        for r in tqdm(self.df):
            wt_emb = torch.load(f'{self.emb_dir}/{r["WT_name"]}_{r["pos"]}{r["wt_seq"][r["pos"] - 1]}.pt')
            mut_emb = torch.load(f'{self.emb_dir}/{r["WT_name"]}_{r["pos"]}{r["aa_mut"]}.pt')

            wt_embs.append(wt_emb)
            mut_embs.append(mut_emb)
        
        self.wt_embs = np.stack(wt_embs)
        self.mut_embs = np.stack(mut_embs)

    def __len__(self):
        return len(self.df)

    def __getitem__(self, idx):
        r = self.df[idx]

        wt_emb = self.wt_embs[idx]
        mut_emb = self.mut_embs[idx]

        # wt_emb = torch.load(f'{self.emb_dir}/{r["WT_name"]}_{r["pos"]}{r["wt_seq"][r["pos"] - 1]}.pt')
        # mut_emb = torch.load(f'{self.emb_dir}/{r["WT_name"]}_{r["pos"]}{r["aa_mut"]}.pt')

        return {
            'wt_emb': wt_emb,
            'mut_emb': mut_emb,
            'label': torch.tensor([r['ddG_ML']]).float(),
        }

if __name__ == '__main__':
    import pandas as pd

    df = pd.read_csv('../data/mega.train.csv')
    train = MegaDataset(df)

    train_loader = DataLoader(train, batch_size=8, shuffle=False)
    for batch in train_loader:
        print(batch['wt_emb'].shape)
        print(batch['mut_emb'].shape)
        print(batch['label'].shape)
        break